==> fred-backend-messenger_0.4.0.orig.tar.gz: fred-backend-messenger-0.4.0/messenger/models/message.py <==
"""Base classes for all message types."""
from typing import Any, Dict, NoReturn, Tuple
from uuid import uuid4

from sqlalchemy import Boolean, Column, Enum, SmallInteger, Text, UniqueConstraint
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.sql import expression, func
from sqlalchemy_utc import UtcDateTime
from sqlalchemy_utils import JSONType, UUIDType
from typing_extensions import Literal

from ..constants import ArchiveMessageStatus, MessageStatus, ReferenceType


def _not_implemented(*args: Any) -> NoReturn:
    raise NotImplementedError


class BaseMessageMixin:
    """Abstract mixin with basic fields shared by all message types."""

    # Envelope attributes
    uuid = property(_not_implemented, _not_implemented)
    create_datetime = property(_not_implemented, _not_implemented)
    update_datetime = Column(UtcDateTime(timezone=True))
    send_datetime = Column(UtcDateTime(timezone=True))
    attempts = Column(SmallInteger(), nullable=False, server_default=expression.text('0'))
    status = property(_not_implemented, _not_implemented)
    type = Column(Text())
    references = property(_not_implemented, _not_implemented)

    # Message attributes
    recipient = property(_not_implemented, _not_implemented)

    # Send attributes
    service = Column(Text(), doc="An alias of a service that sent the message.")

    # Delivery attributes
    delivery_datetime = Column(UtcDateTime(timezone=True))
    delivery_info = Column(JSONType(), doc="Selected information about delivery message.")

    # Archive attribute
    archive = property(_not_implemented, _not_implemented)

    def __repr__(self) -> str:
        return '<{} {} to {}>'.format(type(self).__name__, self.uuid, self.recipient)


class BaseTemplateMessageMixin(BaseMessageMixin):
    """A mixin with basic fields for messages with templates."""

    # Message attributes
    body_template = Column(Text(), nullable=False, doc="Name of the body template.")
    body_template_uuid = Column(UUIDType(binary=False))
    context = Column(JSONType(), doc="Context for all message templates.")


class MessageMixin(BaseMessageMixin):
    """A mixin with basic fields for a regular message object."""

    __table_args__ = {'sqlite_autoincrement': True}  # type: Dict[str, Any]

    # Envelope attributes
    uuid = Column(UUIDType(binary=False), nullable=False, unique=True, default=uuid4)
    create_datetime = Column(UtcDateTime(timezone=True), nullable=False, server_default=func.now())
    # Archive has a different set of statuses
    status = Column(Enum(MessageStatus, create_constraint=True), nullable=False, index=True)

    # Archive attribute
    archive = Column(Boolean(name='archive', create_constraint=True), nullable=False, server_default=expression.true())


class TemplateMessageMixin(MessageMixin, BaseTemplateMessageMixin):
    """A mixin with basic fields for a regular message object with templates."""


class ArchiveMessageMixin(BaseMessageMixin):
    """A mixin with basic fields for an archived message object.

    Also defines partitioning on postgres.
    """

    __table_args__ = {'postgresql_partition_by': 'RANGE (create_datetime)'}

    # Envelope attributes
    # Partitioning prevents unique constraints without partition key.
    uuid = Column(UUIDType(binary=False), nullable=False, default=uuid4, index=True)
    # Partitioning requires partition key to be a part of primary key.
    create_datetime = Column(UtcDateTime(timezone=True), nullable=False, server_default=func.now(), primary_key=True)
    # Archive has a different set of statuses
    status = Column(Enum(ArchiveMessageStatus, create_constraint=True), nullable=False, index=True)

    @property
    def archive(self) -> Literal[True]:  # type: ignore[override]
        """Return True, because the message is already archived."""
        return True


class TemplateArchiveMessageMixin(ArchiveMessageMixin, BaseTemplateMessageMixin):
    """A mixin with basic fields for an archived message object with templates."""


class ReferenceMixin:
    """A mixin with basic fields for a message reference."""

    @declared_attr
    def __table_args__(cls) -> Tuple:
        return (
            UniqueConstraint('value', 'type', 'message_id'),
        )

    message_id = property(_not_implemented, _not_implemented)
    type = Column(Enum(ReferenceType, create_constraint=True), nullable=False)
    value = Column(Text, nullable=False)


class ArchiveReferenceMixin(ReferenceMixin):
    """A mixin with basic fields for an archived message reference.

    Also defines partitioning on postgres.
    """

    @declared_attr
    def __table_args__(cls) -> Tuple:
        return (
            UniqueConstraint('value', 'type', 'message_id', 'create_datetime'),
            {'postgresql_partition_by': 'RANGE (create_datetime)'},
        )

    # We want the partitions constructed in the same way as the messages.
    # Partitioning requires partition key to be a part of primary key.
    create_datetime = Column(UtcDateTime(timezone=True), nullable=False, primary_key=True)

==> fred-backend-messenger_0.5.0.orig.tar.gz: fred-backend-messenger-0.5.0/messenger/models/message.py <==
"""Base classes for all message types."""
from typing import Any, Dict, NoReturn, Optional, Tuple, cast
from uuid import uuid4

from sqlalchemy import Boolean, Column, Enum, SmallInteger, Text, UniqueConstraint
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.sql import expression, func
from sqlalchemy_utc import UtcDateTime
from sqlalchemy_utils import JSONType, UUIDType
from typing_extensions import Literal

from ..constants import ArchiveMessageStatus, MessageStatus, ReferenceType


def _not_implemented(*args: Any) -> NoReturn:
    raise NotImplementedError


class BaseMessageMixin:
    """Abstract mixin with basic fields shared by all message types."""

    # Envelope attributes
    uuid = property(_not_implemented, _not_implemented)
    create_datetime = property(_not_implemented, _not_implemented)
    update_datetime = Column(UtcDateTime(timezone=True))
    send_datetime = Column(UtcDateTime(timezone=True))
    attempts = Column(SmallInteger(), nullable=False, server_default=expression.text('0'))
    status = property(_not_implemented, _not_implemented)
    type = Column(Text())
    references = property(_not_implemented, _not_implemented)

    # Message attributes
    recipient = property(_not_implemented, _not_implemented)

    # Send attributes
    service = Column(Text(), doc="An alias of a service that sent the message.")

    # Delivery attributes
    delivery_datetime = Column(UtcDateTime(timezone=True))
    delivery_info = Column(JSONType(), doc="Selected information about delivery message.")

    # Archive attribute
    archive = property(_not_implemented, _not_implemented)

    def __repr__(self) -> str:
        return '<{} {} to {}>'.format(type(self).__name__, self.uuid, self.recipient)


class BaseTemplateMessageMixin(BaseMessageMixin):
    """A mixin with basic fields for messages with templates."""

    # Message attributes
    body_template = Column(Text(), nullable=False, doc="Name of the body template.")
    body_template_uuid = Column(UUIDType(binary=False))
    context = Column(JSONType(), doc="Context for all message templates.")


class MessageMixin(BaseMessageMixin):
    """A mixin with basic fields for a regular message object."""

    __table_args__ = {'sqlite_autoincrement': True}  # type: Dict[str, Any]

    # Envelope attributes
    uuid = Column(UUIDType(binary=False), nullable=False, unique=True, default=uuid4)
    create_datetime = Column(UtcDateTime(timezone=True), nullable=False, server_default=func.now())
    # Archive has a different set of statuses
    status = Column(Enum(MessageStatus, create_constraint=True), nullable=False, index=True)

    # Archive attribute
    archive = Column(Boolean(name='archive', create_constraint=True), nullable=False, server_default=expression.true())


class TemplateMessageMixin(MessageMixin, BaseTemplateMessageMixin):
    """A mixin with basic fields for a regular message object with templates."""


class ArchiveMessageMixin(BaseMessageMixin):
    """A mixin with basic fields for an archived message object.

    Also defines partitioning on postgres.
    """

    __table_args__ = {'postgresql_partition_by': 'RANGE (create_datetime)'}

    # Envelope attributes
    # Partitioning prevents unique constraints without partition key.
    uuid = Column(UUIDType(binary=False), nullable=False, default=uuid4, index=True)
    # Partitioning requires partition key to be a part of primary key.
    create_datetime = Column(UtcDateTime(timezone=True), nullable=False, server_default=func.now(), primary_key=True)
    # Archive has a different set of statuses
    status = Column(Enum(ArchiveMessageStatus, create_constraint=True), nullable=False, index=True)

    @property
    def archive(self) -> Literal[True]:  # type: ignore[override]
        """Return True, because the message is already archived."""
        return True


class TemplateArchiveMessageMixin(ArchiveMessageMixin, BaseTemplateMessageMixin):
    """A mixin with basic fields for an archived message object with templates."""


class ReferenceMixin:
    """A mixin with basic fields for a message reference."""

    @declared_attr
    def __table_args__(cls) -> Tuple:
        return (
            UniqueConstraint('type', 'value', 'message_id'),
            UniqueConstraint('old_type', 'value', 'message_id'),
        )

    message_id = property(_not_implemented, _not_implemented)
    _old_type = Column("old_type", Enum(ReferenceType, create_constraint=True), nullable=True)
    _new_type = Column("type", Text, nullable=False)
    value = Column(Text, nullable=False)

    @hybrid_property
    def type(self) -> str:
        """Return reference type."""
        return cast(str, self._new_type or self._old_type)

    @type.setter  # type: ignore[no-redef]  # mypy doesn't understand hybrid_property
    def type(self, value: str) -> None:
        """Set reference type."""
        self._new_type = value

    @type.expression  # type: ignore[no-redef]  # mypy doesn't understand hybrid_property
    def type(cls):
        # In SQL just fallback to the new column.
        return cls._new_type

    @property
    def old_type(self) -> Optional[ReferenceType]:
        """Return old-style reference type."""
        if self._old_type:
            return cast(ReferenceType, self._old_type)

        try:
            return ReferenceType(self._new_type)
        except ValueError:
            return None


class ArchiveReferenceMixin(ReferenceMixin):
    """A mixin with basic fields for an archived message reference.

    Also defines partitioning on postgres.
    """

    @declared_attr
    def __table_args__(cls) -> Tuple:
        return (
            UniqueConstraint('type', 'value', 'message_id', 'create_datetime'),
            UniqueConstraint('old_type', 'value', 'message_id', 'create_datetime'),
            {'postgresql_partition_by': 'RANGE (create_datetime)'},
        )

    # We want the partitions constructed in the same way as the messages.
    # Partitioning requires partition key to be a part of primary key.
    create_datetime = Column(UtcDateTime(timezone=True), nullable=False, primary_key=True)
